--Löve2D pathfinding tutorial by BrotsagtMist.
--To be used as main.lua.
--Controls: The arrow keys control your _player_, the space bar generates a new path and lets the _enemy_ move, r places you at random.
--this is just a reference for easy understanding, a real life game can be improved. For example you should not generate a new path for each step your enemy does. 
--while working nice on small maps the performance on huge maps is terrible. Create the path once and then walk it until you hit the end instead.

--First lets define our map.
Map={
{1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,},
{0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,2,0,},--2 is invisibe
{0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,},
{0,0,0,1,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,},
{0,0,0,0,0,1,0,0,0,0,0,0,1,0,0,0,0,0,0,0,0,0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,},
{0,0,0,0,0,1,0,0,0,0,0,1,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,0,0,0,0,0,0,1,0,0,},
{0,0,0,0,0,1,0,0,0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,0,0,0,0,0,1,0,1,0,},
{0,0,0,0,0,1,0,0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,0,0,0,0,1,0,0,0,0,},
{0,0,0,0,0,1,0,0,1,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,1,0,0,0,0,0,},
{0,0,0,0,0,1,0,0,0,0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,0,0,0,},
{0,0,0,0,0,1,1,1,1,1,1,1,0,0,0,0,0,0,1,1,0,0,0,0,0,0,0,0,0,0,1,0,0,0,0,0,},
{0,0,0,0,1,0,0,0,0,0,0,1,0,0,0,0,0,0,1,0,1,0,0,1,0,0,0,0,0,0,0,1,0,0,0,0,},
{0,0,0,1,0,0,0,0,0,0,0,1,0,0,0,0,0,0,1,0,1,0,1,0,0,0,0,0,0,0,0,0,1,0,0,0,},
{0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,0,1,1,0,0,0,0,0,0,0,0,0,0,0,1,0,0,},
{0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,},
{0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,},
}
--It is a table containing another round of tables, containing single numbers that represent each tile. Walkable floor tiles contain an "0" while "1" is rendered as wall.
--Note that a trailing extra comma in it is ignored by Lua which makes copy pasting these lines easier.

--Lets put our map size in extra variables for the sole purpose of being cleaner.
YTiles=#Map 
XTiles=#Map[1]
-- # is the size symbol in Lua, note that this only works for tables which contain ongoing numbers as key, strings will be ignored. 
--!!!! A single tile is accessed by Map[Y][X]. Dont get confused by this reversed order.

--Lets define the size of our tiles in pixel so that the game spans the whole window.
X,Y=love.graphics.getDimensions() --Getting window size, defining two variables at the same time.
XTilesize=math.floor(X/XTiles) --math floor rounds down to an integer. It usually looks nicer if you avoid floats in drawing operations.
YTilesize=math.floor(Y/YTiles)

--Position the player and an enemy at random. We make this a function so we can shuffle later.
--This may place you inside a wall, we dont care.
function Place() --create
 Player={math.random(XTiles),math.random(YTiles)}
 Enemy={math.random(XTiles),math.random(YTiles)}
end 
Place() --and call
--!!!!!math.random does NOT generate a random number. It is more of a list of numbers that is worked through as you call it. 
--You need to set a position within this list to simulate a bit of randomness using for example the current time as a seed.
--Or just use love.math.random instead which seeds automatically.

--The draw loop
function love.draw()
  love.graphics.setColor(0,1,0 )--set wall colours
 --lets traverse through the entire map and draw an rectangle each time we hit an 1. 
 --!!!!!This is NOT very performant, in a real game its better to draw this once to a picture and use this for each frame, read up on "canvas".
 for y=1,YTiles do
  for x=1,XTiles do
   if Map[y][x]==1 then
    love.graphics.rectangle( "fill", XTilesize *x, YTilesize * y, -XTilesize, -YTilesize ) --Note how the Tilesize is negative.
   end                                                                                     --We start painting the rectangle at the lower right corner this time.
  end                                                                                      --This is a con of a start index of 1, but there are numerous ways around it, this is one. 
 end
 --Further this is the only time ever we check for "1" specifically, so we can place other numbers in our map to create either invisible walls or to add walls of different colour here.
   
 --Player
 love.graphics.setColor(0,0,1 )
 love.graphics.rectangle( "fill", XTilesize * Player[1], YTilesize * Player[2], -XTilesize, -YTilesize ) 
 --Enemy
 love.graphics.setColor(1,0,0 )
 love.graphics.rectangle( "fill", XTilesize * Enemy[1], YTilesize * Enemy[2], -XTilesize, -YTilesize ) 
 --draw path 
 love.graphics.setColor(1,1,0 )
 local tail=Enemy[3] -- Our resulting path format is {x,y,{x,y,{x,y,{and so on}}}} thus we can traverse the entire path using this while loop.
 while tail do
  love.graphics.circle( "line", tail[1]*XTilesize-XTilesize/2,tail[2]*YTilesize-XTilesize/2 ,XTilesize/2  )
  tail=tail[3]
 end
end

--Key input
function love.keypressed(k)
 if k=="r" then Place() --shuffle
 elseif k=="up" then --player control
  if Player[2]>1 and Map[Player[2]-1][Player[1]]==0 then  Player[2]=Player[2]-1 end
 elseif k=="down" then 
  if Player[2]<YTiles and Map[Player[2]+1][Player[1]]==0 then  Player[2]=Player[2]+1 end
--Translation: If the players Y coordinate is smaller than the lower border of the map AND the tile under the player is walkable we increase our Y position by one
--Checking the border before the tile is important or we end up indexing a nil value.
 elseif k=="left" then 
  if Map[Player[2]][Player[1]-1]==0 then  Player[1]=Player[1]-1 end
-- we do not need to check for borders this time since a nil is interpreted as wall anyway.
 elseif k=="right" then 
  if Map[Player[2]][Player[1]+1]==0 then  Player[1]=Player[1]+1 end
 elseif k=="space" then -- Call pathfinder, update enemy location
  Enemy= Pathfinder(Player[1],Player[2],Enemy[1],Enemy[2])
  --Often it makes sense to reverse the search, eg when multiple enemy paths need to be created at once, thus we call the pathfinding with the player as position and the enemy as target.
 end
end

--The actual pathfinding process
--The method is a simple bruteforcing and simply touches all tiles until we find our target, in this tutorial target and goal are switched.
--First we need a copy of the entire map for trackeeping.
--Second we create a list of points/nodes to start from, starting with one point in this case, if you start with several points we will end up at the nearest.
--Third we enter a main loop that is to keep active until all pathfinding attempts hit a dead end.
--Fourth for each of our nodes we check in all four directions if we can create a new node, that is, a walkable tile that no other node has been using so far. 
--The current node needs to claim it by deleting it from the trackkeeping map.
--We add the info of where the node came from to the new node, this is only needed for visualising the path.
--Last we break the loop as soon as the target tile was stepped on, signaled by its disappearance from the trackkeeping map, this ensures the shortest possible route is returned 
--If the target was not found we simply return the previous state, so no change was made at all.
function Pathfinder(xs,ys,xe,ye) --xs="X start", xe= "X end" and so on
--Copy the entire map, we keep it local this time since its only used one time.
--This map is used for keeping track of which tiles we already checked out during the traverse, thus the table is garbaged after usage.
 local shadowmap={}
 for y=1,YTiles do
  shadowmap[y]={}
  for x=1,XTiles do
   shadowmap[y][x]=Map[y][x]
  end
 end
 local paths={{xs,ys}} --original starting point
 shadowmap[ys][xs]=nil --we started here, no need to check here
 local newpaths={}--follow up nodes
 while #paths>0 do --this loop breaks once none of the paths generated a valid follow up path
  for k,v in ipairs(paths) do
   if shadowmap[v[2]-1] and shadowmap[v[2]-1][v[1]]==0 then --up
    shadowmap[v[2]-1][v[1]]=nil --delete the tiles we already stepped on
    if not shadowmap[ye][xe] then return v end --we break as soon as we reached our target, thus the returned path is ensured to be the shortest one possible.
    table.insert(newpaths,{v[1],v[2]-1,v}) --add the next tile to the follow up pile
   end 
   if shadowmap[v[2]][v[1]+1]==0 then  --right
    shadowmap[v[2]][v[1]+1]=nil 
    if not shadowmap[ye][xe] then return v end
    table.insert(newpaths,{v[1]+1,v[2],v})
   end 
   if shadowmap[v[2]+1] and shadowmap[v[2]+1][v[1]]==0 then  --down
    shadowmap[v[2]+1][v[1]]=nil 
    if not shadowmap[ye][xe] then return v  end
    table.insert(newpaths,{v[1],v[2]+1,v})
   end 
   if shadowmap[v[2]][v[1]-1]==0 then  --left 
    shadowmap[v[2]][v[1]-1]=nil 
    if not shadowmap[ye][xe] then return v end
    table.insert(newpaths,{v[1]-1,v[2],v}) 
   end 
  end
  paths=newpaths --make follow up nodes the new current ones
  newpaths={}    --and make fresh follow ups
 end
 return Enemy --if no path could be found we simply return the current position
end
--Additional notes:
--This generates paths fast enough to work on huge maps within a single frame. If you call it each frame for real tho you deserve to be hit with a sledgehammer.
--There are numerous tricks to improve performance, for example creating the trackkeeping map together with the nodes thus keeping not accessible parts of the map out of the ram.
--Or checking against multiple target points at once to generate the path for multiple enemies at once in one go.
--But, pathfinding will likely never be the bottleneck of your game, in this very script the most cpu usage comes from the draw part and not the path.
--Which means you can safely ignore improving this and spend your time on more complicated problems.
